@@ -20,7 +20,8 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include "glib.h"
+#include "config.h"
+
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/wait.h>
@@ -34,6 +35,8 @@
 #include <sys/select.h>
 #endif /* HAVE_SYS_SELECT_H */
 
+#include "glib.h"
+
 #include "glibintl.h"
 
 static gint g_execute (const gchar  *file,
@@ -55,7 +58,7 @@ static gboolean fork_exec_with_pipes (gboolean              intermediate_child,
                                       gboolean              file_and_argv_zero,
                                       GSpawnChildSetupFunc  child_setup,
                                       gpointer              user_data,
-                                      gint                 *child_pid,
+                                      GPid                 *child_pid,
                                       gint                 *standard_input,
                                       gint                 *standard_output,
                                       gint                 *standard_error,
@@ -76,7 +79,7 @@ g_spawn_error_quark (void)
  * @argv: child's argument vector
  * @envp: child's environment, or %NULL to inherit parent's
  * @flags: flags from #GSpawnFlags
- * @child_setup: function to run in the child just before <function>exec()</function>
+ * @child_setup: function to run in the child just before exec()
  * @user_data: user data for @child_setup
  * @child_pid: return location for child process ID, or %NULL
  * @error: return location for error
@@ -93,7 +96,7 @@ g_spawn_async (const gchar          *working_directory,
                GSpawnFlags           flags,
                GSpawnChildSetupFunc  child_setup,
                gpointer              user_data,
-               gint                 *child_pid,
+               GPid                 *child_pid,
                GError              **error)
 {
   g_return_val_if_fail (argv != NULL, FALSE);
@@ -117,8 +120,13 @@ close_and_invalidate (gint *fd)
 {
   gint ret;
 
-  ret = close (*fd);
-  *fd = -1;
+  if (*fd < 0)
+    return -1;
+  else
+    {
+      ret = close (*fd);
+      *fd = -1;
+    }
 
   return ret;
 }
@@ -140,7 +148,7 @@ read_data (GString *str,
 
  again:
   
-  bytes = read (fd, &buf, 4096);
+  bytes = read (fd, buf, 4096);
 
   if (bytes == 0)
     return READ_EOF;
@@ -171,19 +179,18 @@ read_data (GString *str,
  * @argv: child's argument vector
  * @envp: child's environment, or %NULL to inherit parent's
  * @flags: flags from #GSpawnFlags
- * @child_setup: function to run in the child just before <function>exec()</function>
+ * @child_setup: function to run in the child just before exec()
  * @user_data: user data for @child_setup
  * @standard_output: return location for child output 
  * @standard_error: return location for child error messages
- * @exit_status: child exit status, as returned by <function>waitpid()</function>
+ * @exit_status: child exit status, as returned by waitpid()
  * @error: return location for error
  *
  * Executes a child synchronously (waits for the child to exit before returning).
  * All output from the child is stored in @standard_output and @standard_error,
  * if those parameters are non-%NULL. If @exit_status is non-%NULL, the exit 
  * status of the child is stored there as it would be returned by 
- * <function>waitpid()</function>; standard UNIX macros such as 
- * <function>WIFEXITED()</function> and <function>WEXITSTATUS()</function> 
+ * waitpid(); standard UNIX macros such as WIFEXITED() and WEXITSTATUS() 
  * must be used to evaluate the exit status. If an error occurs, no data is 
  * returned in @standard_output, @standard_error, or @exit_status.
  * 
@@ -206,7 +213,7 @@ g_spawn_sync (const gchar          *working_directory,
 {
   gint outpipe = -1;
   gint errpipe = -1;
-  gint pid;
+  GPid pid;
   fd_set fds;
   gint ret;
   GString *outstr = NULL;
@@ -255,12 +262,12 @@ g_spawn_sync (const gchar          *working_directory,
 
   if (outpipe >= 0)
     {
-      outstr = g_string_new ("");
+      outstr = g_string_new (NULL);
     }
       
   if (errpipe >= 0)
     {
-      errstr = g_string_new ("");
+      errstr = g_string_new (NULL);
     }
 
   /* Read data until we get EOF on both pipes. */
@@ -407,7 +414,7 @@ g_spawn_sync (const gchar          *working_directory,
  * @argv: child's argument vector
  * @envp: child's environment, or %NULL to inherit parent's
  * @flags: flags from #GSpawnFlags
- * @child_setup: function to run in the child just before <function>exec()</function>
+ * @child_setup: function to run in the child just before exec()
  * @user_data: user data for @child_setup
  * @child_pid: return location for child process ID, or %NULL
  * @standard_input: return location for file descriptor to write to child's stdin, or %NULL
@@ -424,54 +431,100 @@ g_spawn_sync (const gchar          *working_directory,
  * the program must be a full path; the <envar>PATH</envar> shell variable 
  * will only be searched if you pass the %G_SPAWN_SEARCH_PATH flag.
  *
+ * On Windows, the low-level child process creation API
+ * (CreateProcess())doesn't use argument vectors,
+ * but a command line. The C runtime library's
+ * <function>spawn*()</function> family of functions (which
+ * g_spawn_async_with_pipes() eventually calls) paste the argument
+ * vector elements into a command line, and the C runtime startup code
+ * does a corresponding recostruction of an argument vector from the
+ * command line, to be passed to main(). Complications arise when you have
+ * argument vector elements that contain spaces of double quotes. The
+ * <function>spawn*()</function> functions don't do any quoting or
+ * escaping, but on the other hand the startup code does do unquoting
+ * and unescaping in order to enable receiving arguments with embedded
+ * spaces or double quotes. To work around this asymmetry,
+ * g_spawn_async_with_pipes() will do quoting and escaping on argument
+ * vector elements that need it before calling the C runtime
+ * spawn() function.
+ *
  * @envp is a %NULL-terminated array of strings, where each string
  * has the form <literal>KEY=VALUE</literal>. This will become
  * the child's environment. If @envp is %NULL, the child inherits its
  * parent's environment.
  *
  * @flags should be the bitwise OR of any flags you want to affect the
- * function's behavior. The %G_SPAWN_DO_NOT_REAP_CHILD means that the
- * child will not be automatically reaped; you must call 
- * <function>waitpid()</function> or handle %SIGCHLD yourself, or the 
- * child will become a zombie.
+ * function's behavior. On Unix, the %G_SPAWN_DO_NOT_REAP_CHILD means
+ * that the child will not be automatically reaped; you must call
+ * waitpid() or handle %SIGCHLD yourself, or the
+ * child will become a zombie. On Windows, the flag means that a
+ * handle to the child will be returned @child_pid. You must call
+ * CloseHandle() on it eventually (or exit the
+ * process), or the child processs will continue to take up some table
+ * space even after its death. Quite similar to zombies on Unix,
+ * actually.
+ *
  * %G_SPAWN_LEAVE_DESCRIPTORS_OPEN means that the parent's open file
  * descriptors will be inherited by the child; otherwise all
  * descriptors except stdin/stdout/stderr will be closed before
- * calling <function>exec()</function> in the child. %G_SPAWN_SEARCH_PATH 
+ * calling exec() in the child. %G_SPAWN_SEARCH_PATH 
  * means that <literal>argv[0]</literal> need not be an absolute path, it
  * will be looked for in the user's <envar>PATH</envar>. 
- * %G_SPAWN_STDOUT_TO_DEV_NULL means that the child's standad output will 
+ * %G_SPAWN_STDOUT_TO_DEV_NULL means that the child's standard output will 
  * be discarded, instead of going to the same location as the parent's 
- * standard output.
+ * standard output. If you use this flag, @standard_output must be %NULL.
  * %G_SPAWN_STDERR_TO_DEV_NULL means that the child's standard error
- * will be discarded. %G_SPAWN_CHILD_INHERITS_STDIN means that
- * the child will inherit the parent's standard input (by default,
- * the child's standard input is attached to /dev/null).
+ * will be discarded, instead of going to the same location as the parent's
+ * standard error. If you use this flag, @standard_error must be %NULL.
+ * %G_SPAWN_CHILD_INHERITS_STDIN means that the child will inherit the parent's
+ * standard input (by default, the child's standard input is attached to
+ * /dev/null). If you use this flag, @standard_input must be %NULL.
  * %G_SPAWN_FILE_AND_ARGV_ZERO means that the first element of @argv is
  * the file to execute, while the remaining elements are the
  * actual argument vector to pass to the file. Normally
  * g_spawn_async_with_pipes() uses @argv[0] as the file to execute, and
  * passes all of @argv to the child.
  *
- * @child_setup and @user_data are a function and user data to be
- * called in the child after GLib has performed all the setup it plans
- * to perform (including creating pipes, closing file descriptors,
- * etc.) but before calling <function>exec()</function>. That is, 
- * @child_setup is called just before calling <function>exec()</function> 
- * in the child. Obviously actions taken in this function will only affect 
- * the child, not the parent. 
+ * @child_setup and @user_data are a function and user data. On POSIX
+ * platforms, the function is called in the child after GLib has
+ * performed all the setup it plans to perform (including creating
+ * pipes, closing file descriptors, etc.) but before calling
+ * exec(). That is, @child_setup is called just
+ * before calling exec() in the child. Obviously
+ * actions taken in this function will only affect the child, not the
+ * parent. On Windows, there is no separate fork() and exec()
+ * functionality. Child processes are created and run right away with
+ * one API call, CreateProcess(). @child_setup is
+ * called in the parent process just before creating the child
+ * process. You should carefully consider what you do in @child_setup
+ * if you intend your software to be portable to Windows.
  *
- * If non-%NULL, @child_pid will be filled with the child's process
- * ID. You can use the process ID to send signals to the child, or
- * to <function>waitpid()</function> if you specified the 
- * %G_SPAWN_DO_NOT_REAP_CHILD flag.
+ * If non-%NULL, @child_pid will on Unix be filled with the child's
+ * process ID. You can use the process ID to send signals to the
+ * child, or to waitpid() if you specified the
+ * %G_SPAWN_DO_NOT_REAP_CHILD flag. On Windows, @child_pid will be
+ * filled with a handle to the child process only if you specified the
+ * %G_SPAWN_DO_NOT_REAP_CHILD flag. You can then access the child
+ * process using the Win32 API, for example wait for its termination
+ * with the <function>WaitFor*()</function> functions, or examine its
+ * exit code with GetExitCodeProcess(). You should close the handle 
+ * with CloseHandle() when you no longer need it.
  *
  * If non-%NULL, the @standard_input, @standard_output, @standard_error
  * locations will be filled with file descriptors for writing to the child's
  * standard input or reading from its standard output or standard error.
  * The caller of g_spawn_async_with_pipes() must close these file descriptors
- * when they are no longer in use. If these parameters are %NULL, the
- * corresponding pipe won't be created.
+ * when they are no longer in use. If these parameters are %NULL, the corresponding
+ * pipe won't be created.
+ *
+ * If @standard_input is NULL, the child's standard input is attached to /dev/null
+ * unless %G_SPAWN_CHILD_INHERITS_STDIN is set.
+ *
+ * If @standard_error is NULL, the child's standard error goes to the same location
+ * as the parent's standard error unless %G_SPAWN_STDERR_TO_DEV_NULL is set.
+ *
+ * If @standard_output is NULL, the child's standard output goes to the same location
+ * as the parent's standard output unless %G_SPAWN_STDOUT_TO_DEV_NULL is set.
  *
  * @error can be %NULL to ignore errors, or non-%NULL to report errors.
  * If an error is set, the function returns %FALSE. Errors
@@ -482,6 +535,9 @@ g_spawn_sync (const gchar          *working_directory,
  *
  * If an error occurs, @child_pid, @standard_input, @standard_output,
  * and @standard_error will not be filled with valid values.
+ *
+ * If @child_pid is not %NULL and an error does not occur then the returned
+ * pid must be closed using g_spawn_close_pid().
  * 
  * Return value: %TRUE on success, %FALSE if an error was set
  **/
@@ -492,7 +548,7 @@ g_spawn_async_with_pipes (const gchar          *working_directory,
                           GSpawnFlags           flags,
                           GSpawnChildSetupFunc  child_setup,
                           gpointer              user_data,
-                          gint                 *child_pid,
+                          GPid                 *child_pid,
                           gint                 *standard_input,
                           gint                 *standard_output,
                           gint                 *standard_error,
@@ -543,6 +599,14 @@ g_spawn_async_with_pipes (const gchar          *working_directory,
  * appropriate. Possible errors are those from g_spawn_sync() and those
  * from g_shell_parse_argv().
  * 
+ * On Windows, please note the implications of g_shell_parse_argv()
+ * parsing @command_line. Space is a separator, and backslashes are
+ * special. Thus you cannot simply pass a @command_line containing
+ * canonical Windows paths, like "c:\\program files\\app\\app.exe", as
+ * the backslashes will be eaten, and the space will act as a
+ * separator. You need to enclose such paths with single quotes, like
+ * "'c:\\program files\\app\\app.exe' 'e:\\folder\\argument.txt'".
+ *
  * Return value: %TRUE on success, %FALSE if an error was set
  **/
 gboolean
@@ -590,6 +654,8 @@ g_spawn_command_line_sync (const gchar  *command_line,
  * consider using g_spawn_async() directly if appropriate. Possible
  * errors are those from g_shell_parse_argv() and g_spawn_async().
  * 
+ * The same concerns on Windows apply as for g_spawn_command_line_sync().
+ *
  * Return value: %TRUE on success, %FALSE if error is set.
  **/
 gboolean
@@ -726,13 +792,36 @@ exec_err_to_g_error (gint en)
     }
 }
 
+static gssize
+write_all (gint fd, gconstpointer vbuf, gsize to_write)
+{
+  gchar *buf = (gchar *) vbuf;
+  
+  while (to_write > 0)
+    {
+      gssize count = write (fd, buf, to_write);
+      if (count < 0)
+        {
+          if (errno != EINTR)
+            return FALSE;
+        }
+      else
+        {
+          to_write -= count;
+          buf += count;
+        }
+    }
+  
+  return TRUE;
+}
+
 static void
 write_err_and_exit (gint fd, gint msg)
 {
   gint en = errno;
   
-  write (fd, &msg, sizeof(msg));
-  write (fd, &en, sizeof(en));
+  write_all (fd, &msg, sizeof(msg));
+  write_all (fd, &en, sizeof(en));
   
   _exit (1);
 }
@@ -938,13 +1027,13 @@ fork_exec_with_pipes (gboolean              intermediate_child,
                       gboolean              file_and_argv_zero,
                       GSpawnChildSetupFunc  child_setup,
                       gpointer              user_data,
-                      gint                 *child_pid,
+                      GPid                 *child_pid,
                       gint                 *standard_input,
                       gint                 *standard_output,
                       gint                 *standard_error,
                       GError              **error)     
 {
-  gint pid;
+  GPid pid = -1;
   gint stdin_pipe[2] = { -1, -1 };
   gint stdout_pipe[2] = { -1, -1 };
   gint stderr_pipe[2] = { -1, -1 };
@@ -1007,15 +1096,15 @@ fork_exec_with_pipes (gboolean              intermediate_child,
            * is to exit, so we can waitpid() it immediately.
            * Then the grandchild will not become a zombie.
            */
-          gint grandchild_pid;
+          GPid grandchild_pid;
 
           grandchild_pid = fork ();
 
           if (grandchild_pid < 0)
             {
               /* report -1 as child PID */
-              write (child_pid_report_pipe[1], &grandchild_pid,
-                     sizeof(grandchild_pid));
+              write_all (child_pid_report_pipe[1], &grandchild_pid,
+                         sizeof(grandchild_pid));
               
               write_err_and_exit (child_err_report_pipe[1],
                                   CHILD_FORK_FAILED);              
@@ -1040,7 +1129,7 @@ fork_exec_with_pipes (gboolean              intermediate_child,
             }
           else
             {
-              write (child_pid_report_pipe[1], &grandchild_pid, sizeof(grandchild_pid));
+              write_all (child_pid_report_pipe[1], &grandchild_pid, sizeof(grandchild_pid));
               close_and_invalidate (&child_pid_report_pipe[1]);
               
               _exit (0);
@@ -1185,7 +1274,9 @@ fork_exec_with_pipes (gboolean              intermediate_child,
         }
       
       /* Success against all odds! return the information */
-      
+      close_and_invalidate (&child_err_report_pipe[0]);
+      close_and_invalidate (&child_pid_report_pipe[0]);
+ 
       if (child_pid)
         *child_pid = pid;
 
@@ -1200,6 +1291,26 @@ fork_exec_with_pipes (gboolean              intermediate_child,
     }
 
  cleanup_and_fail:
+
+  /* There was an error from the Child, reap the child to avoid it being
+     a zombie.
+   */
+
+  if (pid > 0)
+  {
+    wait_failed:
+     if (waitpid (pid, NULL, 0) < 0)
+       {
+          if (errno == EINTR)
+            goto wait_failed;
+          else if (errno == ECHILD)
+            ; /* do nothing, child already reaped */
+          else
+            g_warning ("waitpid() should not fail in "
+                       "'fork_exec_with_pipes'");
+       }
+   }
+
   close_and_invalidate (&child_err_report_pipe[0]);
   close_and_invalidate (&child_err_report_pipe[1]);
   close_and_invalidate (&child_pid_report_pipe[0]);
@@ -1409,3 +1520,17 @@ g_execute (const gchar *file,
   /* Return the error from the last attempt (probably ENOENT).  */
   return -1;
 }
+
+/**
+ * g_spawn_close_pid:
+ * @pid: The process identifier to close
+ *
+ * On some platforms, notably WIN32, the #GPid type represents a resource
+ * which must be closed to prevent resource leaking. g_spawn_close_pid()
+ * is provided for this purpose. It should be used on all platforms, even
+ * though it doesn't do anything under UNIX.
+ **/
+void
+g_spawn_close_pid (GPid pid)
+{
+}
